/**
 * 郵便番号・住所サジェスト システム
 *
 * @version   1.00.00
 * @copyright (c) 2019 Step One co,.ltd.
 * @license   Commercial License
 * @see       https://www.steponecoltd.com/product/p2as/
 */
{
  const VERSION = '1.00.00';
  const SEARCH_PHP_FILE_PATH = 'AUTO';

  const U = undefined, N = null, T = true, F = false,
        W = window, D = W.document;
  const $ = 'querySelectorAll';

  const MULTI_CSS = [ '#s1-p2as-popup { position: absolute; opacity: 0; margin: 0; padding: 0; border: 1px solid #666; background-color: #fff; z-index: 65535; }',
                      '#s1-p2as-popup div { display: inline-block; }',
                      '#s1-p2as-popup div:nth-child(2), #s1-p2as-popup div:nth-child(3) { margin-left: 1px; border-left: 1px dotted #888; padding-left: 1px; }',
                      '#s1-p2as-popup span { display: block; padding: 1px 14px 2px 14px; font-size: 14px; color: #000; cursor: default; }',
                      '#s1-p2as-popup span:hover { background-color: #000088; color: #fff; }' ];
  let multiInited = F;
  let currentPostCode = N;


  //初期化関数
  const init = () => {
    let searchPath = SEARCH_PHP_FILE_PATH,
        sElms, elms, t;

    //パスの自動取得
    if (searchPath === 'AUTO') {
      t = D[$]('SCRIPT[src*="p2as.js"]');
      if (t[0]) {
        searchPath = t[0].src.replace(/[^\/]+$/, '') + 'p2as-search.php';
      } else {
        error('p2as-search.php detection failed.');
      }
    }

    //セットの読み込み
    sElms = D[$]('*[data-p2as="set"]');
    if (!sElms.length) {    //セットがない場合は BODY にしておく
      sElms = D[$]('BODY');
    }

    if (!sElms.forEach) {   //※polyfill
      sElms = _arrayFrom(sElms);
    }

    //セット毎に初期化
    sElms.forEach((sElm) => {
      let conf = { type: N, searchPath: searchPath, fieldMap: {} },
          fMap = conf.fieldMap,
          type, fElms, key, pcElms, val, r;

      //フィールドを取得
      fElms = sElm[$]('*[data-p2as]');

      if (!fElms.forEach) {   //※polyfill
        fElms = _arrayFrom(fElms);
      }

      fElms.forEach((fElm) => {
        key = (fElm.dataset.p2as || '').replace(/ /g, '');

        if (!fMap[key]) {
          fMap[key] = [];
        }
        fMap[key].push(fElm);

        if (key.indexOf('multi') !== -1) {
          type = 'multi';
        }
      });
      if (fMap.jigyosyo || fMap.company) {
        type = type || 'jigyosyo';
      }
      conf.type = type || '1to1';

      //変更関数を定義
      const change = () => {
        let tVal;
        val = '';                                   //フィールドから値を読み込む
        pcElms.forEach((pcElm) => {
          val += pcElm.value;
        });

        if (r = val.match(/^(\d{3})\-?(\d{4})$/)) { //郵便番号形式の場合
          tVal = (r[1] + r[2]) * 1;

          //同じ場合はスルー
          if (currentPostCode === tVal) {
            return;
          }
          currentPostCode = tVal;

          conf.value = tVal;
          conf.origin = val;
          request(conf);
        }
      };

      //郵便番号のフィールドにイベントの付与
      pcElms = fMap['postcode'];
      if (pcElms) {
        pcElms.forEach((pcElm) => {
          pcElm.addEventListener('keyup', change, false);
          pcElm.addEventListener('paste', () => {
            setTimeout(change);
          }, false);
        });
      } else {
        error('postcode field not exist.');
      }
    });
  };

  //街区／地番版 初期化関数
  const multiInit = () => {
    let styElm, css, i, l;

    //スタイル シートを追加
    styElm = D.createElement('style');
    styElm.type = 'text/css';
    D.getElementsByTagName('HEAD')[0].appendChild(styElm);
    css = D.styleSheets[D.styleSheets.length - 1];
    for (i = 0, l = MULTI_CSS.length; i < l; i++) {
      css.insertRule(MULTI_CSS[i], i);
    }

    multiInited = T;
  };

  //リクエスト関数
  let request;
  if (window.fetch) {             //※Firefox, Chrome, Edge
    request = (conf) => {
      let src, fOpt, p;

      src  = conf.searchPath + 
             '?postcode=' + (conf.value + '') + '&type=' + conf.type;
      fOpt = { method: 'GET' };

      //fetch で取得
      p = fetch(src, fOpt);
      p.then((res) => {
        if (res.ok) {
          return res.text();
        } else {
          error('network error. (' + res.status + ' ' + res.statusText + ')');
          return;
        }
      }).then((body) => {   //ボディ取得
        response(body, conf)
      }).catch((err) => {   //エラー (ネットワーク エラー)
        error('network error. (' + err + ')');
      });
    };
  } else {                        //※IE
    request = function(conf) {
      let src, fOpt, req, p;

      src = conf.searchPath +
            '?postcode=' + (conf.value + '') + '&type=' + conf.type;

      req = new XMLHttpRequest();
      req.open('GET', src, T);
      req.addEventListener('readystatechange', () => {
        if (req.readyState === 4) {
          response(req.responseText, conf);
        }
      }, false);
      req.send(null);
    };
  }

  //レスポンス関数
  const response = (body, conf) => {
    let fMap = conf.fieldMap,
        cVal = '', res, multi, r;

    //値が検索した時のものと入れ替わっているかチェック
    conf.fieldMap['postcode'].forEach((pcElm) => {
      cVal += pcElm.value;
    });
    if (cVal !== conf.origin) {
      return;
    }

    //レスポンス値を解析
    if (body.indexOf('{') !== -1 && (r = body.match(/\{(.+)\}/))) { //マルチ抽出
      multi = r[1];
      body = body.replace(r[0], '');
    }
    res = body.match(/^(\d+)([^,]+),([^,]+)(?:,(.+))?$/);
    if (!res) {
      if (body.includes('data no match')) {   //データなし
        return;
      } else {                                //それ以外のエラー
        error('response parse failed.');
        return;
      }
    }
    res = res.slice(1);       //先頭は除外
    if (multi) {
      res.push(multi);
    }
    res = res.map((v, i) => {   //念のためサニタイズしておく
      return (v || '').replace(/[<>"']/g, '');
    });

    setField(conf, res);
  };

  //フィールドにセット
  const setField = (conf, res) => {
    let fMap = conf.fieldMap,
        [ prefCode, pref, address, company, multi ] = res,
        oElms, prefPat, t, i, il, mb;

    //街区／地番版の場合、データを解析
    if (multi) {
      multi = parseMulti(multi, T);
    }

    //都道府県+市区町村
    t = fMap['prefecture+address'] || fMap['prefecture+address+multi'];
    if (t && t[0]) {
      if (!t[0].value) {
        t[0].value = pref + address;
      }
      mb = pref + address;
    }

    //都道府県
    t = fMap['prefecture'];
    if (t && t[0]) {
      if (t[0].type === 'text' && !t[0].value) {    //テキストフィールド
        t[0].value = pref;
      } else if (t[0].tagName === 'SELECT') {       //セレクトボックス
        oElms = t[0][$]('option');
        prefPat = [ pref, pref.replace(/(?:都|道|府|県)$/, ''),
                    prefCode, prefCode.replace(/^0/, '') ];
        for (i = 0, il = oElms.length; i < il; i++) {
          if (prefPat.some((ptn) => {
                return (oElms[i].value === ptn ? T : F);
              })) {
            oElms[i].selected = T;
            break;
          }
        }
      }
    }

    //市区町村
    t = fMap['address'] || fMap['address+multi'];
    if (t && t[0] && address) {
      if (!t[0].value) {
        t[0].value = address;
      }
      mb = address;
    }

    //事業所
    t = fMap['jigyosyo'] || fMap['company'];
    if (t && t[0] && !t[0].value && company) {
      t[0].value = company;
    }

    //街区／地番版の場合、ポップアップを表示
    if (multi) {
      if (!multiInited) {
        multiInit();
      }
      t = (fMap['prefecture+address+multi'] || fMap['address+multi']);
      openMultiPopup(t[0], mb, multi);
    }

  };

  //街区／地番版データを解析
  const parseMulti = (val, first = false) => {
    let buf = [], ret = {}, list = {}, func, i, l, t, bef = null;

    //初回のみ * を展開   例: 3*6 => 3,4,5,6
    if (first && val.includes('*')) {
      val = val.replace(/(\d+)\*(\d+)/g, (s, m1, m2) => {
        for (i = m1 * 1, l = m2 * 1, t = []; i <= l; i++) {
          t.push(i);
        }
        return t.join(',');
      });
    }

    //入れ子をバッファに挿入
    do {
      bef = val;
      val = val.replace(/\{([^\{\}]+)\}/g, (s, m) => {
        buf.push(m);
        return '\t' + (buf.length - 1);
      });

      if (val === bef) {
        break;
      }
    } while (true);

    //展開してツリー構成を構築
    func = (val, pref) => {
      let ret = [], k;
      val.split(',').forEach((v) => {
        let p = v.indexOf('\t');

        if (p === -1) {   //下位層なし
          ret.push(v);
          k = pref + '|' + v;
          v = null;
        } else {          //下位層あり
          t = v.substr(0, p);
          ret.push(t);
          k = pref + '|' + t
          v = func(buf[v.substr(p + 1)], k);
        }
        k = k.replace(/(\d)\|(\d)/, '$1-$2').replace(/\|/g, '');

        list[k] = v;
      });
      return ret;
    };
    list[''] = func(val, '');

    return list;
  };

  //街区／地番版データを解析
  const openMultiPopup = (tElm, base, data) => {
    let dElm, pElm, css = '', fSize, val, crtData, crtDataGroup = [], pos,
        evFunc, len, div, limit, t, t1, t2, w, h, i, l, k, br;

    //値から一覧を取得
    val = tElm.value;
    crtData = data[val.replace(new RegExp('^' + base), '')];
    if (val.indexOf(base) === -1 || !crtData || !crtData.length) {
      return;
    }

    //ダミーエレメントを生成しスタイルをコピー
    dElm = D.createElement('DIV');
    t1 = D.defaultView.getComputedStyle(tElm, '');
    t2 = D.defaultView.getComputedStyle(dElm, '');
    for (i = 0, l = t1.length; i < l; i++) {    //※ for of 使用不可
      k = t1[i];
      if (t1[k] !== t2[k]) {
        css += k + ': ' + t1[k] + '; ';
      }
    }
    dElm.style.cssText = css;
    fSize = t1.fontSize;

    //値、プロパティ、スタイル等をセットしてターゲットの隣に生成
    val = val.replace(/</g, '&lt;').replace(/>/g, '&gt;')
             .replace(/\n/g, '<br>');
    dElm.innerHTML = val + '<span>.</span>';
    dElm.style.position   = 'absolute';
    dElm.style.overflow   = 'hidden';
    dElm.style.whiteSpace = "pre";
    dElm.style.opacity    = '0';
    tElm.parentNode.insertBefore(dElm, tElm.nextSibling);

    //調整して取得
    dElm.scrollLeft = tElm.scrollLeft;
    pos = { left: dElm.firstElementChild.offsetLeft, top: 0 };
    pos.left -= dElm.scrollLeft;

    t = tElm.getBoundingClientRect();   //絶対座標に変換
    pos.left += t.left + W.pageXOffset;
    pos.top  += t.top  + W.pageYOffset;

    //ダミーエレメントを削除
    dElm.parentNode.removeChild(dElm);

    //ポップアップを作成
    pElm = D.createElement('DIV');
    pElm.id = 's1-p2as-popup';
    pElm.style.fontSize = fSize;

    //表示する一覧を作成して追加
    len = crtData.length;
    div = (len >= 20 ? 3 : (len >= 10 ? 2 : 1));
    limit = Math.ceil(len / div);
    for (i = 0, l = len; i < l; i+= limit) {
      crtDataGroup.push(crtData.slice(i, i + limit));
    }
    t1 = '';
    for (t2 of crtDataGroup) {
      for (i = 0, l = limit - t2.length, br = ''; i < l; i++) {
        br += '<br>';
      }
      t1 += '<div><span>' + t2.join('</span><span>') + '</span>' + br + '</div>';
    }
    pElm.innerHTML = t1;
    D.getElementsByTagName('BODY')[0].appendChild(pElm);

    //サイズを取得し、位置を決定、表示
    w = pElm.offsetWidth;
    h = pElm.offsetHeight;

    t1 =  D.documentElement.clientHeight;
    t2 = W.pageYOffset;

    if (pos.top + tElm.offsetHeight + h - W.pageYOffset < t1 || //下に表示
        pos.top - h < t2 ) {
        pos.top += tElm.offsetHeight - 1;
    } else {                                                    //上に表示
        pos.top -= h - 1;
    }

    pElm.style.left = pos.left + 'px';
    pElm.style.top  = pos.top  + 'px';
    pElm.style.opacity = '1';


    //イベントをセット
    evFunc = (ev) => {
      let elm = ev.target, pntElm = elm.parentNode.parentNode, t;
      if (pntElm && pntElm.id === 's1-p2as-popup') { //ポップアップ内をクリック
        tElm.value = (tElm.value + '|' + elm.innerHTML).replace(/[<>]/g, '')
                               .replace(/(\d)\|(\d)/, '$1-$2').replace(/\|/, '');

        setTimeout(() => { openMultiPopup(tElm, base, data); }, 350);
      } else {                                        //ポップアップ外をクリック
        pntElm = D.getElementById('s1-p2as-popup');
      }

      pntElm.parentNode.removeChild(pntElm);  //ポップアップを削除
      W.removeEventListener('click', evFunc); //イベントを削除
    };
    setTimeout(() => { W.addEventListener('click', evFunc); }, 100);


  };


  //エラー関数
  const error = (mes) => {
    console.error('[ PostcodeToAddressAutoComplete ERROR ]\n' + mes);
  };


  W.addEventListener('DOMContentLoaded', init, { once: T });
}
